const double valueOfPi = 3.1415926535;

#include "Arduino.h"
#include <Adafruit_GFX.h>       // установка с сайта github -> https://github.com/adafruit/Adafruit-GFX-Library
#include <Adafruit_ST7735.h>    // установка с сайта github -> https://github.com/adafruit/Adafruit-ST7735-Library
#include <SPI.h>

/*
    ----------------- настройки для клавиатуры --------------------
    pin_0 - 0-й разряд, к нему подключена кнопка запуска калибровки
    pin_1 - 1-й разряд, к нему подключена кнопка запуска переключения измерения L или C
    Когда нажата кнопка "переключения измерения L или C" сканкод клавиатуры будет 0b00000001
    Когда нажата кнопка "запуска калибровки" сканкод клавиатуры будет 0b00000010
*/
#define KEYBOARD_PIN_0        GPIO_NUM_2
#define KEYBOARD_PIN_1        GPIO_NUM_4
#define SUCCESS_HITS_NUMBER 6
#define KEYBOARD_MASK 0b00000011
int keyboardHitCounter = 0;
uint8_t keyboardPreviousCode = KEYBOARD_MASK;
bool buttonPressedFlag = false;
bool waitButtonRelease = false;

#include <WiFi.h>
#include <AsyncTCP.h>           // установка с сайта github -> https://github.com/me-no-dev/AsyncTCP 
#include <ESPAsyncWebServer.h>  // установка с сайта github -> https://github.com/me-no-dev/ESPAsyncWebServer 
#include <AsyncElegantOTA.h>    // установка через менеджер библиотек: http://librarymanager/All#AsyncElegantOTA
AsyncWebServer server(80);

/* авторизационные парметры для OTA (беспроводного обновления прошивки) */
const char* ota_username = "firmware";   // имя пользователя для беспроводного обновления прошивки
const char* ota_password = "firmware";   // пароль для беспроводного обновления прошивки

/* параметры собственной беспроводной точки доступа */
String ap_ssid = "L-Meter";       // имя беспроводной сети, вещаемое точкой доступа
String ap_password = "12345678";  // пароль для подключения к точке доступа


#define relayControlPin       GPIO_NUM_27   // пин, который управляет реле переключения измерения L или C
#define FREQ_METTER_IN        GPIO_NUM_34   // пин, на который подается измеряемая частота

/* параметры и пины, используемые для подключения графического дисплея */
#define TFT_RST    16
#define TFT_DC     17
#define TFT_CS     5
Adafruit_ST7735 * tft;

// емкость "рефренсного" конденсатора (тут нужна максимальная точность!)
double ref_capacitance = 10.185E-9;

/* емкость рефренсной катушки индуктивности (тут точность не особо критична, т.к.
  значение переписывается программой после выполнения процедуры калибровки) */
double ref_inductance = 78.6E-6;

/* остальные переменные */
uint8_t currentMode = 0;
uint8_t rememberCurrentMode = 0;
uint8_t currentDisplay = 0 ;
double capacitance, valueToDisplay;
String component, unit1, unit1000 , unit1000000;
bool rememberRelayState = false;


volatile bool timer0IntFlag = false;
volatile bool oneSecondFlag = false;
volatile int oneSecondCounter = 0;

volatile uint32_t freqInputCounter = 0;
volatile uint32_t currentFrequency = 0;
uint32_t calibrationFrequency = 0;
/*
   Процедура обработки прерывания 0-го таймера.
*/
hw_timer_t * timer0 = NULL;
portMUX_TYPE timer0Mux = portMUX_INITIALIZER_UNLOCKED;
void IRAM_ATTR onTimer0() {
  portENTER_CRITICAL_ISR(&timer0Mux);
  uint32_t tempValue = freqInputCounter;             // временно запоминаем значение счетчика частоты
  oneSecondCounter = (oneSecondCounter + 1) % 100 ;  // 100 срабатываний таймера 0 равны 1-й секунде
  if (oneSecondCounter == 0) {                       // если было 100 срабатываний, то
    currentFrequency = tempValue;                    // значение счетчика частоты записываем как текущую частоту
    freqInputCounter = 0;                            // сбрасываем счетчик частоты
    oneSecondFlag = true;                            // ставим флаг "одна секунда"
  }
  timer0IntFlag = true;                              // ставим флаг срабатывания таймера 0
  portEXIT_CRITICAL_ISR(&timer0Mux);
}

/* Обработчик прерывания по входному сигналу частотометра */
portMUX_TYPE ext_Int_2 = portMUX_INITIALIZER_UNLOCKED;
void IRAM_ATTR handleExtInterrupt() {
  portENTER_CRITICAL_ISR(&ext_Int_2);
  freqInputCounter ++;                               // просто увеличиваем счетчик частоты на единицу
  portEXIT_CRITICAL_ISR(&ext_Int_2);
}

void setup()
{
  /* настройка выхода ESP32, используемого для управления реле */
  pinMode( relayControlPin, OUTPUT );
  digitalWrite( relayControlPin, LOW );
  /* настройка входов, к которым подключены кнопки управления */
  pinMode(KEYBOARD_PIN_0, INPUT_PULLUP);
  pinMode(KEYBOARD_PIN_1, INPUT_PULLUP);
  /* инициализация дисплея */
  tft = new Adafruit_ST7735(TFT_CS, TFT_DC, TFT_RST);
  tft->initR( INITR_BLACKTAB );
  tft->setRotation( 1 );

  delay( 1000 );

  /* запуск точки доступа Wi-Fi*/
  WiFi.mode(WIFI_MODE_AP);
  WiFi.softAP(ap_ssid.c_str(), ap_password.c_str());

  /* переадресация запроса с индексной страницы веб-сервера на страницу обновления прошивки */
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->redirect("/update");
  });

  /* Запуск ElegantOTA ( => возможность обновления прошивки "по воздуху") */
  AsyncElegantOTA.begin(&server, ota_username, ota_password);

  /* запускаем сам веб-сервер */
  server.begin();

  /* Запуск таймера 0, который будует давать прерывание раз в 10 миллисекунд */
  /* https://techtutorialsx.com/2017/10/07/esp32-arduino-timer-interrupts/ */
  timer0 = timerBegin(0, 80, true); // тут: 0 - номер таймера, 80 - прескаллер, true - постоянная работа
  timerAttachInterrupt(timer0, &onTimer0, true); // подключаем прерывание
  timerAlarmWrite(timer0, 10000, true);  // ставим делитель такой, чтобы прерывание происходило каждые 10 мсек
  timerAlarmEnable(timer0); // включаем сам таймер

  /* настройка входа ESP32, используемого для измерения частоты генератора */
  pinMode(FREQ_METTER_IN, INPUT);
  attachInterrupt(digitalPinToInterrupt(FREQ_METTER_IN), handleExtInterrupt, RISING);

}

void loop()
{
  /*
     Каждые 10 миллисекунд (по флагу таймера 0) сканируем клавиатуру.
     Если кнопка была нажата, вызываем процедуру "анализза", после
     чего сбрасываем флаг(и)
  */
  if (timer0IntFlag) {
    readKeyboardData();
    if (buttonPressedFlag) {
      analizeKeyboard();
      buttonPressedFlag = false;
    }
    timer0IntFlag = false;
  }

  /*
     Раз в секунду обновляем показания дисплея. Если запущена "калибровка"
     (текущий режим больше 2), то еще и вызываем ее процедуру.
  */
  if (oneSecondFlag) {
    if (currentMode > 2) {
      exaclyCalibration();
    }
    displayInformation();
    oneSecondFlag = false;
  }
}

/*
   "анализ клавиатуры" - вызов нужной процедуры 
   в зависимости от полученного сканкода
*/
void analizeKeyboard() {
  switch (keyboardPreviousCode) {
    case 0b00000010:
      startOfCalibration();
      break;
    case 0b00000001:
      switchMode();
      displayInformation();
      break;
  }
}

void displayInformation() {
  switch (currentMode) {
    /* отображение показаний индуктивности */
    case 0:
      valueToDisplay = getValueOf(ref_capacitance, ref_inductance, currentFrequency, 1E6);
      unit1 = " uH";
      unit1000 = " mH";
      unit1000000 = " H";
      component = "inductor";
      if (currentFrequency < 1000) {
        showNotConnected();
        break;
      }
      if (currentDisplay != currentMode ) {
        tft->fillScreen(ST77XX_BLACK);
        tft->drawRect(2, 2, 156, 124, 0xAD35);
        currentDisplay = currentMode;
      }
      showExaclyValue();
      break;
    /* отображение показаний емкости */
    case 1:
      valueToDisplay = getValueOf(ref_inductance, ref_capacitance, currentFrequency, 1E12);
      unit1 = " pF";
      unit1000 = " nF";
      unit1000000 = " uF";
      component = "capasitor";
      if (currentFrequency > 200000 or valueToDisplay < 0.0 ) {
        showNotConnected();
        break;
      }
      if (currentDisplay != currentMode ) {
        tft->fillScreen(ST77XX_BLACK);
        tft->drawRect(2, 2, 156, 124, 0xAD35);
        currentDisplay = currentMode;
      }
      showExaclyValue();
      break;
  }
}

void showExaclyValue() {
  if ( valueToDisplay >= 1000000 ) {
    drawTexxt( 31, 35, 100, ST77XX_BLACK, ST77XX_GREEN, 2, String( valueToDisplay / 1000000 ) + unit1000000 );
    drawTexxt( 20, 64, 120, ST77XX_BLACK, ST77XX_YELLOW, 1, "Freq : " + String( currentFrequency ) + " Hz" );
  } else if ( valueToDisplay >= 1000 ) {
    drawTexxt( 31, 35, 100, ST77XX_BLACK, ST77XX_GREEN, 2, String( valueToDisplay / 1000 ) + unit1000 );
    drawTexxt( 20, 64, 120, ST77XX_BLACK, ST77XX_YELLOW, 1, "Freq : " + String( currentFrequency ) + " Hz" );
  } else {
    drawTexxt( 31, 36, 100, ST77XX_BLACK, ST77XX_GREEN, 2, String( valueToDisplay ) +  unit1 );
    drawTexxt( 20, 64, 120, ST77XX_BLACK, ST77XX_YELLOW, 1, "Freq : " + String( currentFrequency ) + " Hz" );
  }
  drawTexxt( 50, 85, 100, ST77XX_BLACK, ST77XX_RED, 1, "L1=" + String( ref_inductance * 1E6) + " uH" );
}

void showNotConnected() {
  if (currentDisplay != 2) {
    tft->fillScreen(ST77XX_RED);
    tft->drawRect(2, 2, 156, 124, ST77XX_BLACK);
    drawTexxt( 35, 46, 120, ST77XX_RED, ST77XX_YELLOW, 1, "Connect" );
    drawTexxt( 35, 60, 120, ST77XX_RED, ST77XX_YELLOW, 1, component );
    drawTexxt( 35, 74, 120, ST77XX_RED, ST77XX_YELLOW, 1, "to measure" );
    currentDisplay = 2;
  }
}

void showWait() {
  if (currentDisplay != 3) {
    tft->fillScreen(ST77XX_RED);
    tft->drawRect(2, 2, 156, 124, ST77XX_BLACK);
    drawTexxt( 50, 54, 100, ST77XX_RED, ST77XX_BLACK, 3, "WAIT" );
    currentDisplay = 3;
  }
}

/* процедура вывода на дисплей одной строки текста */
void drawTexxt( int16_t x, int16_t y, int16_t maxTextWidth, int bgColor, int txtColor, uint8_t fontSize, String text )
{
  int pos = (160 - ( text.length() * 6 * fontSize )) / 2; // вычисляем стартовую позицию для текста в строке (с учетом размещения текста по центру)
  tft->fillRect( 3, y - 2 , 154, ( 8 * fontSize ) + 3, bgColor ); // "закрашиваем" строку цветом фона
  tft->setCursor( pos, y ); // устанавливаем курсор на позицию старта строки
  tft->setTextColor( txtColor ); // устанавливаем цвет текста
  tft->setTextWrap( false );  // выключаем перенос сроки
  tft->setTextSize(fontSize); // устанавливаем размер шрифта
  tft->print( text ); // выводим текст
}

/* процедура переключения режима измерения с L на C и обратно */
void switchMode() {
  switch (currentMode) {
    case 0:                                    // если был режим "0" (измерение индуктивности)...
      currentMode = 1;                         // ...то включаем режим "1" (измерение емкости)
      digitalWrite( relayControlPin, HIGH );   // ... и подаем сигнал для включения реле
      break;
    case 1:                                    // если был режим "1" (измерение емкости)...
      currentMode = 0;                         // ...то включаем режим "0" (измерение индуктивности)
      digitalWrite( relayControlPin, LOW );    // ... и выключаем реле
      break;
  }
}

/* запуск процедуры калибровки прибора */
void startOfCalibration() {
  rememberCurrentMode = currentMode;                  // запоминаем режим, из которого запускалась калибровка
  rememberRelayState = digitalRead(relayControlPin);  // запоминаем состояние выхода управления реле "R или C"
  currentMode = 3;                                    // переключаем текущий режим на старт "калибровки" (3) и
  exaclyCalibration();                                // ... вызываем реальную процедуру калибровки
}

/* непосредственно сама процедура калибровки */
void exaclyCalibration() {
  switch (currentMode) {
    case 3:
      digitalWrite( relayControlPin, HIGH );
      showWait();
      currentMode++;
      break;
    case 4:
      calibrationFrequency = 0;
      currentMode++;
      break;
    case 5:
      calibrationFrequency += currentFrequency;
      currentMode++;
      break;
    case 6:
      calibrationFrequency += currentFrequency;
      calibrationFrequency = calibrationFrequency / 2 ;
      currentMode++;
      break;
    case 7:
      calibrationFrequency += currentFrequency;
      calibrationFrequency = calibrationFrequency / 2 ;
      ref_inductance = getValueOf(ref_capacitance, 0.0, calibrationFrequency, 1.0);
      digitalWrite( relayControlPin, rememberRelayState );
      currentMode = rememberCurrentMode;
      break;
  }
}

/* процедура вычисления одного из номиналов контура по измеренной частоте и рефренсному номиналу второго */
float getValueOf(double anotherValue, double refrenceValue, uint32_t mesuredFreq, double multiplaxer) {
  float toReturn = (1.0 / ( 4.0 * mesuredFreq * mesuredFreq * anotherValue * valueOfPi * valueOfPi )) - refrenceValue;
  toReturn *= multiplaxer;
  return toReturn;
}

/* сканирование клавиатуры, защита от дребезга, задержка после отпускания */
void readKeyboardData() {
  /* считываем текущее состояние пинов клавиатуры */
  byte NewScanCode = (gpio_get_level(KEYBOARD_PIN_1) << 1) | gpio_get_level(KEYBOARD_PIN_0);
  /* отбрасываем лишние биты по маске (на всякий случай) */
  NewScanCode &= KEYBOARD_MASK;
  /* эта проверка определяет задержку между отпусканием предыдущей кнопки и нажатием следующей */
  if (keyboardPreviousCode == KEYBOARD_MASK && keyboardHitCounter < (10 * SUCCESS_HITS_NUMBER)) {
    keyboardHitCounter++ ;
    return;
  }
  /* если старый и новый сканкоды совпали, */
  if (NewScanCode == keyboardPreviousCode) {
    /* .. просто увеличиваем счетчик совпадений сканкодов */
    keyboardHitCounter++ ;
  } else {
    keyboardPreviousCode = NewScanCode;  // "новый" сканкод становится "старым"
    keyboardHitCounter = 0;              // счетчик совпадений сбрасывается
    buttonPressedFlag = false;           // флаг "была нажата кнопка" сбрасывается (чисто на всякий случай)
  }
  /* если число повторов равно заданному (кнопка была нажата дольше времени подавления дребезга) */
  if ( keyboardHitCounter == SUCCESS_HITS_NUMBER) {
    /* устанавливаем флаг "была нажата кнопка" */
    buttonPressedFlag = true;
  }
  /* этот "костыль" не даст переполнится счетчику повторов, сколько бы не держали кнопку нажатой */
  if (keyboardHitCounter > SUCCESS_HITS_NUMBER) {
    keyboardHitCounter --;
  }
}
